@@ -62,6 +62,30 @@
 (defvar ement-room-prism-minimum-contrast)
 (defvar ement-room-unread-only-counts-notifications)
 
+;;;; Function declarations
+
+;; Instead of using top-level `declare-function' forms (which can easily become obsolete
+;; if not kept with the code that needs them), this allows the use of `(declare (function
+;; ...))' forms in each function definition, so that if a function is moved or removed,
+;; the `declare-function' goes with it.
+
+;; TODO: Propose this upstream.
+
+(eval-and-compile
+  (defun ement--byte-run--declare-function (_name _args &rest values)
+    "Return a `declare-function' form with VALUES.
+Allows the use of a form like:
+
+  (declare (function FN FILE ...))
+
+inside of a function definition, effectively keeping its
+`declare-function' form inside the function definition, ensuring
+that stray such forms don't remain if the function is removed."
+    `(declare-function ,@values))
+
+  (cl-pushnew '(function ement--byte-run--declare-function) defun-declarations-alist :test #'equal)
+  (cl-pushnew '(function ement--byte-run--declare-function) macro-declarations-alist :test #'equal))
+
 ;;;; Compatibility
 
 ;; These workarounds should be removed when they aren't needed.
@@ -500,7 +524,12 @@ otherwise use current room."
                          ;; alignment problems.
                          (spec (format "%%-%ss %%s" name-width)))
               (save-excursion
-                (insert "\"" (propertize (or display-name canonical-alias room-id) 'face 'font-lock-doc-face) "\"" " is a room "
+                (insert "\"" (propertize (or display-name canonical-alias room-id) 'face 'font-lock-doc-face) "\"" " is a "
+                        (propertize (if (ement--space-p room)
+                                        "space"
+                                      "room")
+                                    'face 'font-lock-type-face)
+                        " "
                         (propertize (pcase status
                                       ('invite "invited")
                                       ('join "joined")
@@ -836,7 +865,7 @@ USER is an `ement-user' struct."
 ;;               ;; 3.
 ;;               (cl-third (servers-by-population-in room))))))))
 
-(defun ement--room-space-p (room)
+(defun ement--space-p (room)
   "Return non-nil if ROOM is a space."
   (equal "m.space" (ement-room-type room)))
 
@@ -981,21 +1010,13 @@ period, anywhere in the body."
     (pcase-let* (((cl-struct ement-room members) room)
                  (regexp (rx (or bos bow (1+ blank))
                              (or (seq (group
-                                       ;; Group 1: full MXID or @-prefixed displayname.
+                                       ;; Group 1: full @-prefixed MXID.
                                        "@" (group
-                                            ;; Group 2: MXID username or displayname.
-                                            ;; NOTE: We special-case the question mark,
-                                            ;; period, and comma so they can be used after
-                                            ;; a displayname or MXID in a sentence (if
-                                            ;; they are present in a displayname, too
-                                            ;; bad).
-                                            (1+ (not (any blank "?.,")))))
-                                      ;; NOTE: Including punctuation in this terminator
-                                      ;; means that wild displaynames with non-word
-                                      ;; characters might not get matched, but it's
-                                      ;; necessary so that a mention can be like "What
-                                      ;; about @foobar?"
-                                      (or eow eos (syntax punctuation) (seq ":" (1+ blank))))
+                                            ;; Group 2: displayname.  (NOTE: Does not work
+                                            ;; with displaynames containing spaces.)
+                                            (1+ (seq (optional ".") alnum)))
+                                       (optional ":" (1+ (seq (optional ".") alnum))))
+                                      (or ":" eow eos (syntax punctuation)))
                                  (seq (group
                                        ;; Group 3: MXID username or displayname.
                                        (1+ (not blank)))
@@ -1008,14 +1029,9 @@ period, anywhere in the body."
                         (setf replace-group 1)
                         (format template (match-string 1 body)
                                 (ement--xml-escape-string (ement--user-displayname-in room member))))
-                      (when-let* ((name (or (when (match-string 1 body)
-                                              ;; Found @-prefixed displayname: replace the
-                                              ;; whole thing, including the @, but only
-                                              ;; use the displayname for looking up the user.
+                      (when-let* ((name (or (when (match-string 2 body)
                                               (setf replace-group 1)
                                               (match-string 2 body))
-                                            (prog1 (match-string 2 body)
-                                              (setf replace-group 2))
                                             (prog1 (match-string 3 body)
                                               (setf replace-group 3))))
                                   (members (members-having-displayname name members))
@@ -1228,20 +1244,25 @@ DATA is an unsent message event's data alist."
   "Return non-nil if event A replaces event B.
 That is, if event A replaces B in their
 \"m.relates_to\"/\"m.relations\" and \"m.replace\" metadata."
-  (pcase-let* (((cl-struct ement-event (id a-id)
+  (pcase-let* (((cl-struct ement-event (id a-id) (origin-server-ts a-ts)
                            (content (map ('m.relates_to
                                           (map ('rel_type a-rel-type)
                                                ('event_id a-replaces-event-id))))))
                 a)
-               ((cl-struct ement-event (id b-id)
-                           ;; Not sure why this ends up in the unsigned key, but it does.
-                           (unsigned (map ('m.relations
-                                           (map ('m.replace
-                                                 (map ('event_id b-replaced-by-event-id))))))))
+               ((cl-struct ement-event (id b-id) (origin-server-ts b-ts)
+                           (content (map ('m.relates_to
+                                          (map ('rel_type b-rel-type)
+                                               ('event_id b-replaces-event-id)))
+                                         ('m.relations
+                                          (map ('m.replace
+                                                (map ('event_id b-replaced-by-event-id))))))))
                 b))
-    (or (and (equal "m.replace" a-rel-type)
-             (equal a-replaces-event-id b-id))
-        (equal a-id b-replaced-by-event-id))))
+    (or (equal a-id b-replaced-by-event-id)
+        (and (equal "m.replace" a-rel-type)
+             (or (equal a-replaces-event-id b-id)
+                 (and (equal "m.replace" b-rel-type)
+                      (equal a-replaces-event-id b-replaces-event-id)
+                      (>= a-ts b-ts)))))))
 
 (defun ement--events-equal-p (a b)
   "Return non-nil if events A and B are essentially equal.
