@@ -39,6 +39,23 @@
   "Group Ement rooms with Taxy."
   :group 'ement)
 
+;;;; Mouse commands
+
+;; Since mouse-activated commands must handle mouse events, we define a simple macro to
+;; wrap a command into a mouse-event-accepting one.
+
+(defmacro ement-room-list-define-mouse-command (command)
+  "Define a command that calls COMMAND interactively with point at mouse event.
+COMMAND should be a form that evaluates to a function symbol; if
+a symbol, it should be unquoted.."
+  (let ((docstring (format "Call command `%s' interactively with point at EVENT." command))
+        (name (intern (format "ement-room-list-mouse-%s" command))))
+    `(defun ,name (event)
+       ,docstring
+       (interactive "e")
+       (mouse-set-point event)
+       (call-interactively #',command))))
+
 ;;;; Variables
 
 (declare-function ement-room-toggle-space "ement-room")
@@ -48,10 +65,23 @@
     (define-key map (kbd "RET") #'ement-room-list-RET)
     (define-key map (kbd "SPC") #'ement-room-list-next-unread)
     (define-key map [tab] #'ement-room-list-section-toggle)
-    (define-key map [mouse-1] #'ement-room-list-mouse-1)
+    (define-key map [mouse-1] (ement-room-list-define-mouse-command ement-room-list-RET))
+    (define-key map [mouse-2] (ement-room-list-define-mouse-command ement-room-list-kill-buffer))
     (define-key map (kbd "k") #'ement-room-list-kill-buffer)
     (define-key map (kbd "s") #'ement-room-toggle-space)
-    map))
+    map)
+  "Keymap for `ement-room-list' buffers.
+See also `ement-room-list-button-map'.")
+
+(defvar ement-room-list-button-map
+  ;; This map is needed because some columns are propertized as buttons, which override
+  ;; the main keymap.
+  ;; TODO: Is it possible to adjust the button properties to obviate this map?
+  (let ((map (make-sparse-keymap)))
+    (define-key map [mouse-1] (ement-room-list-define-mouse-command ement-room-list-RET))
+    (define-key map [mouse-2] (ement-room-list-define-mouse-command ement-room-list-kill-buffer))
+    map)
+  "Keymap for buttonized text in `ement-room-list' buffers.")
 
 (defvar ement-room-list-timestamp-colors nil
   "List of colors used for timestamps.
@@ -369,9 +399,10 @@ from recent to non-recent for rooms updated in the past hour.")
              (push 'ement-room-list-invited (map-elt face :inherit)))
             ('leave
              (push 'ement-room-list-left (map-elt face :inherit))))
-          (propertize (ement--button-buttonize display-name #'ement-room-list-mouse-1)
+          (propertize display-name
                       'face face
-                      'mouse-face 'highlight))
+                      'mouse-face 'highlight
+                      'keymap ement-room-list-button-map))
         "")))
 
 (ement-room-list-define-column #("Unread" 0 6 (help-echo "Unread events (Notifications:Highlights)")) (:align 'right)
@@ -516,7 +547,8 @@ After showing it, its window is selected.  The buffer is named
 BUFFER-NAME and is shown with DISPLAY-BUFFER-ACTION; or if
 DISPLAY-BUFFER-ACTION is nil, the buffer is not displayed."
   (interactive)
-  (let (format-table column-sizes window-start)
+  (let ((window-start 0) (window-point 0)
+        format-table column-sizes)
     (cl-labels (;; (heading-face
                 ;;  (depth) (list :inherit (list 'bufler-group (bufler-level-face depth))))
                 (format-item (item) (gethash item format-table))
@@ -584,77 +616,93 @@ DISPLAY-BUFFER-ACTION is nil, the buffer is not displayed."
       ;;   (kill-buffer buffer-name))
       (unless ement-sessions
         (error "Ement: Not connected.  Use `ement-connect' to connect"))
-      (with-current-buffer (get-buffer-create buffer-name)
-        (ement-room-list-mode)
-        (let* ((room-session-vectors
-                (cl-loop for (_id . session) in ement-sessions
-                         append (cl-loop for room in (ement-session-rooms session)
-                                         collect (vector room session))))
-               (taxy (cl-macrolet ((first-item
-                                    (pred) `(lambda (taxy)
-                                              (when (taxy-items taxy)
-                                                (,pred (car (taxy-items taxy))))))
-                                   (name= (name) `(lambda (taxy)
-                                                    (equal ,name (taxy-name taxy)))))
-                       (thread-last
-                         (make-fn
-                          :name "Ement Rooms"
-                          :take (taxy-make-take-function keys ement-room-list-keys))
-                         (taxy-fill room-session-vectors)
-                         (taxy-sort #'> #'item-latest-ts)
-                         (taxy-sort #'t<nil #'item-invited-p)
-                         (taxy-sort #'t<nil #'item-favourite-p)
-                         (taxy-sort #'t>nil #'item-low-priority-p)
-                         (taxy-sort #'t<nil #'item-unread-p)
-                         (taxy-sort #'t<nil #'item-space-p)
-                         ;; Within each taxy, left rooms should be sorted last so that one
-                         ;; can never be the first room in the taxy (unless it's the taxy
-                         ;; of left rooms), which would cause the taxy to be incorrectly
-                         ;; sorted last.
-                         (taxy-sort #'t>nil #'item-left-p)
-                         (taxy-sort* #'string< #'taxy-name)
-                         (taxy-sort* #'> #'taxy-latest-ts)
-                         (taxy-sort* #'t<nil (name= "Buffers"))
-                         (taxy-sort* #'t<nil (first-item item-unread-p))
-                         (taxy-sort* #'t<nil (first-item item-favourite-p))
-                         (taxy-sort* #'t<nil (first-item item-invited-p))
-                         (taxy-sort* #'t>nil (first-item item-space-p))
-                         (taxy-sort* #'t>nil (name= "Low-priority"))
-                         (taxy-sort* #'t>nil (first-item item-left-p)))))
-               (taxy-magit-section-insert-indent-items nil)
-               (inhibit-read-only t)
-               (format-cons (taxy-magit-section-format-items
-                             ement-room-list-columns ement-room-list-column-formatters taxy))
-               (pos (point))
-               (section-ident (when (magit-current-section)
-                                (magit-section-ident (magit-current-section)))))
-          (setf format-table (car format-cons)
-                column-sizes (cdr format-cons)
-                header-line-format (taxy-magit-section-format-header
-                                    column-sizes ement-room-list-column-formatters)
-                window-start (if (get-buffer-window buffer-name)
-                                 (window-start (get-buffer-window buffer-name))
-                               0))
-          (when ement-room-list-visibility-cache
-            (setf magit-section-visibility-cache ement-room-list-visibility-cache))
-          (add-hook 'kill-buffer-hook #'ement-room-list--cache-visibility nil 'local)
-          (delete-all-overlays)
-          (erase-buffer)
-          (save-excursion
-            (taxy-magit-section-insert taxy :items 'first
-              ;; :blank-between-depth bufler-taxy-blank-between-depth
-              :initial-depth 0))
-          (goto-char pos)
-          (when (and section-ident (magit-get-section section-ident))
-            (goto-char (oref (magit-get-section section-ident) start)))))
-      (when display-buffer-action
-        (when-let ((window (display-buffer buffer-name display-buffer-action)))
-          (select-window window)))
-      (when (get-buffer-window buffer-name)
-        (set-window-start (get-buffer-window buffer-name) window-start))
-      ;; NOTE: In order for `bookmark--jump-via' to work properly, the restored buffer
-      ;; must be set as the current buffer, so we have to do this explicitly here.
-      (set-buffer buffer-name))))
+      (if (not (cl-loop for (_id . session) in ement-sessions
+                        thereis (ement-session-rooms session)))
+          (ement-message "No rooms have been joined")
+        (with-current-buffer (get-buffer-create buffer-name)
+          (unless (eq 'ement-room-list-mode major-mode)
+            (ement-room-list-mode))
+          (let* ((room-session-vectors
+                  (cl-loop for (_id . session) in ement-sessions
+                           append (cl-loop for room in (ement-session-rooms session)
+                                           collect (vector room session))))
+                 (taxy (cl-macrolet ((first-item
+                                      (pred) `(lambda (taxy)
+                                                (when (taxy-items taxy)
+                                                  (,pred (car (taxy-items taxy))))))
+                                     (name= (name) `(lambda (taxy)
+                                                      (equal ,name (taxy-name taxy)))))
+                         (thread-last
+                           (make-fn
+                            :name "Ement Rooms"
+                            :take (taxy-make-take-function keys ement-room-list-keys))
+                           (taxy-fill room-session-vectors)
+                           (taxy-sort #'> #'item-latest-ts)
+                           (taxy-sort #'t<nil #'item-invited-p)
+                           (taxy-sort #'t<nil #'item-favourite-p)
+                           (taxy-sort #'t>nil #'item-low-priority-p)
+                           (taxy-sort #'t<nil #'item-unread-p)
+                           (taxy-sort #'t<nil #'item-space-p)
+                           ;; Within each taxy, left rooms should be sorted last so that one
+                           ;; can never be the first room in the taxy (unless it's the taxy
+                           ;; of left rooms), which would cause the taxy to be incorrectly
+                           ;; sorted last.
+                           (taxy-sort #'t>nil #'item-left-p)
+                           (taxy-sort* #'string< #'taxy-name)
+                           (taxy-sort* #'> #'taxy-latest-ts)
+                           (taxy-sort* #'t<nil (name= "Buffers"))
+                           (taxy-sort* #'t<nil (first-item item-unread-p))
+                           (taxy-sort* #'t<nil (first-item item-favourite-p))
+                           (taxy-sort* #'t<nil (first-item item-invited-p))
+                           (taxy-sort* #'t>nil (first-item item-space-p))
+                           (taxy-sort* #'t>nil (name= "Low-priority"))
+                           (taxy-sort* #'t>nil (first-item item-left-p)))))
+                 (taxy-magit-section-insert-indent-items nil)
+                 (inhibit-read-only t)
+                 (format-cons (taxy-magit-section-format-items
+                               ement-room-list-columns ement-room-list-column-formatters taxy))
+                 (pos (point))
+                 (section-ident (when (magit-current-section)
+                                  (magit-section-ident (magit-current-section)))))
+            (setf format-table (car format-cons)
+                  column-sizes (cdr format-cons)
+                  header-line-format (taxy-magit-section-format-header
+                                      column-sizes ement-room-list-column-formatters))
+            (when-let ((window (get-buffer-window (current-buffer))))
+              (setf window-point (window-point window)
+                    window-start (window-start window)))
+            (when ement-room-list-visibility-cache
+              (setf magit-section-visibility-cache ement-room-list-visibility-cache))
+            (add-hook 'kill-buffer-hook #'ement-room-list--cache-visibility nil 'local)
+            ;; Before this point, no changes have been made to the buffer's contents.
+            (delete-all-overlays)
+            (erase-buffer)
+            (save-excursion
+              (taxy-magit-section-insert taxy :items 'first
+                ;; :blank-between-depth bufler-taxy-blank-between-depth
+                :initial-depth 0))
+            (if-let* ((section-ident)
+                      (section (magit-get-section section-ident)))
+                (goto-char (oref section start))
+              (goto-char pos))))
+        (when display-buffer-action
+          (when-let ((window (display-buffer buffer-name display-buffer-action)))
+            (select-window window)))
+        (when-let ((window (get-buffer-window buffer-name)))
+          (set-window-start window window-start)
+          (set-window-point window window-point))
+        ;; FIXME: Despite all this code to save and restore point and window point and
+        ;; window start, when I send a message from the minibuffer, or when I abort
+        ;; sending a message from the minibuffer, point is moved to the beginning of the
+        ;; buffer.  While the minibuffer is open (and the typing messages are being sent
+        ;; to the server, causing it to repeatedly sync), the point stays in the correct
+        ;; place.  I can't find any reason why this happens.  It makes no sense.  And
+        ;; while trying to debug the problem, somehow Emacs got put into an unbreakable,
+        ;; infinite loop twice; even C-g and SIGUSR2 didn't stop it.
+
+        ;; NOTE: In order for `bookmark--jump-via' to work properly, the restored buffer
+        ;; must be set as the current buffer, so we have to do this explicitly here.
+        (set-buffer buffer-name)))))
 
 (cl-defun ement-room-list-side-window (&key (side 'left))
   "Show room list in side window on SIDE.
@@ -692,19 +740,16 @@ left."
       (kill-buffer buffer)
       (ement-room-list-revert))))
 
-(defun ement-room-list-mouse-1 (event)
-  "Call `ement-room-list-RET' at EVENT."
-  (interactive "e")
-  (mouse-set-point event)
-  (call-interactively #'ement-room-list-RET))
-
 (declare-function ement-view-room "ement-room")
 (defun ement-room-list-RET ()
   "View room at point, or cycle section at point."
+  (declare (function ement-view-space "ement-room"))
   (interactive)
   (cl-etypecase (oref (magit-current-section) value)
     (vector (pcase-let ((`[,room ,session] (oref (magit-current-section) value)))
-              (ement-view-room room session)))
+              (if (ement--space-p room)
+                  (ement-view-space room session)
+                (ement-view-room room session))))
     (taxy-magit-section (call-interactively #'ement-room-list-section-toggle))
     (null nil)))
 
@@ -712,19 +757,17 @@ left."
 (defun ement-room-list-next-unread ()
   "Show next unread room."
   (interactive)
-  (unless (button-at (point))
-    (call-interactively #'forward-button))
+  (when (eobp)
+    (goto-char (point-min)))
   (unless (cl-loop with starting-line = (line-number-at-pos)
                    for value = (oref (magit-current-section) value)
-                   for room = (elt value 0)
-                   for session = (elt value 1)
-                   if (ement--room-unread-p room session)
+                   if (and (vectorp value)
+                           (ement--room-unread-p (elt value 0) (elt value 1)))
                    do (progn
-                        (goto-char (button-end (button-at (point))))
-                        (push-button (1- (point)))
+                        (ement-view-room (elt value 0)  (elt value 1))
                         (ement-room-goto-fully-read-marker)
                         (cl-return t))
-                   else do (call-interactively #'forward-button)
+                   else do (forward-line 1)
                    while (> (line-number-at-pos) starting-line))
     ;; No more unread rooms.
     (message "No more unread rooms")))
