@@ -5,8 +5,8 @@
 ;; Author: Adam Porter <adam@alphapapa.net>
 ;; Maintainer: Adam Porter <adam@alphapapa.net>
 ;; URL: https://github.com/alphapapa/ement.el
-;; Version: 0.6
-;; Package-Requires: ((emacs "27.1") (map "2.1") (plz "0.2") (taxy "0.10") (taxy-magit-section "0.12.1") (svg-lib "0.2.5") (transient "0.3.7"))
+;; Version: 0.10
+;; Package-Requires: ((emacs "27.1") (map "2.1") (persist "0.5") (plz "0.6") (taxy "0.10") (taxy-magit-section "0.12.1") (svg-lib "0.2.5") (transient "0.3.7"))
 ;; Keywords: comm
 
 ;; This program is free software; you can redistribute it and/or modify
@@ -184,6 +184,11 @@ handled (e.g. how to be notified)."
   :type 'hook
   :options '(ement-interrupted-sync-message ement-interrupted-sync-warning))
 
+(defcustom ement-sso-server-port 4567
+  "TCP port used for local HTTP server for SSO logins.
+It shouldn't usually be necessary to change this."
+  :type 'integer)
+
 ;;;; Commands
 
 ;;;###autoload
@@ -205,8 +210,7 @@ the port, e.g.
   \"http://localhost:8080\""
   (interactive (if current-prefix-arg
                    ;; Force new session.
-                   (list :user-id (read-string "User ID: ")
-                         :password (read-passwd "Password: "))
+                   (list :user-id (read-string "User ID: "))
                  ;; Use known session.
                  (unless ement-sessions
                    ;; Read sessions from disk.
@@ -215,60 +219,106 @@ the port, e.g.
                      (error (display-warning 'ement (format "Unable to read session data from disk (%s).  Prompting to log in again."
                                                             (error-message-string err))))))
                  (cl-case (length ement-sessions)
-                   (0 (list :user-id (read-string "User ID: ")
-                            :password (read-passwd "Password: ")))
+                   (0 (list :user-id (read-string "User ID: ")))
                    (1 (list :session (cdar ement-sessions)))
                    (otherwise (list :session (ement-complete-session))))))
-  (cl-labels ((new-session
-               () (unless (string-match (rx bos "@" (group (1+ (not (any ":")))) ; Username
-                                            ":" (group (optional (1+ (not (any blank)))))) ; Server name
-                                        user-id)
-                    (user-error "Invalid user ID format: use @USERNAME:SERVER"))
-               (let* ((username (match-string 1 user-id))
-                      (server-name (match-string 2 user-id))
-                      (uri-prefix (or uri-prefix (ement--hostname-uri server-name)))
-                      (user (make-ement-user :id user-id :username username))
-                      (server (make-ement-server :name server-name :uri-prefix uri-prefix))
-                      (transaction-id (ement--initial-transaction-id)))
-                 (make-ement-session :user user :server server :transaction-id transaction-id
-                                     :events (make-hash-table :test #'equal))))
-              (password-login
-               () (pcase-let* (((cl-struct ement-session user device-id initial-device-display-name) session)
-                               ((cl-struct ement-user id) user)
-                               (data (ement-alist "type" "m.login.password"
-                                                  "identifier"
-                                                  (ement-alist "type" "m.id.user"
-                                                               "user" id)
-                                                  "password" password
-                                                  "device_id" device-id
-                                                  "initial_device_display_name" initial-device-display-name)))
-                    ;; TODO: Clear password in callback (if we decide to hold on to it for retrying login timeouts).
-                    (ement-api session "login" :method 'post :data (json-encode data)
-                      :then (apply-partially #'ement--login-callback session))))
-              (flows-callback
-               (data) (if (cl-loop for flow across (map-elt data 'flows)
-                                   thereis (equal (map-elt flow 'type) "m.login.password"))
-                          (progn
-                            (message "Ement: Logging in with password...")
-                            (password-login))
-                        (error "Matrix server doesn't support m.login.password login flow.  Supported flows: %s"
-                               (cl-loop for flow in (map-elt data 'flows)
-                                        collect (map-elt flow 'type))))))
-    (if session
-        ;; Start syncing given session.
-        (let ((user-id (ement-user-id (ement-session-user session))))
-          ;; HACK: If session is already in ement-sessions, this replaces it.  I think that's okay...
-          (setf (alist-get user-id ement-sessions nil nil #'equal) session)
-          (ement--sync session :timeout ement-initial-sync-timeout))
-      ;; Start password login flow.  Prompt for user ID and password
-      ;; if not given (i.e. if not called interactively.)
-      (unless user-id
-        (setf user-id (read-string "User ID: ")))
-      (unless password
-        (setf password (read-passwd (format "Password for %s: " user-id))))
-      (setf session (new-session))
-      (when (ement-api session "login" :then #'flows-callback)
-        (message "Ement: Checking server's login flows...")))))
+  (let (sso-server-process)
+    (cl-labels ((new-session
+                 () (unless (string-match (rx bos "@" (group (1+ (not (any ":")))) ; Username
+                                              ":" (group (optional (1+ (not (any blank)))))) ; Server name
+                                          user-id)
+                      (user-error "Invalid user ID format: use @USERNAME:SERVER"))
+                 (let* ((username (match-string 1 user-id))
+                        (server-name (match-string 2 user-id))
+                        (uri-prefix (or uri-prefix (ement--hostname-uri server-name)))
+                        (user (make-ement-user :id user-id :username username))
+                        (server (make-ement-server :name server-name :uri-prefix uri-prefix))
+                        (transaction-id (ement--initial-transaction-id))
+                        (initial-device-display-name (format "Ement.el: %s@%s"
+                                                             ;; Just to be extra careful:
+                                                             (or user-login-name "[unknown user-login-name]")
+                                                             (or (system-name) "[unknown system-name]")))
+                        (device-id (secure-hash 'sha256 initial-device-display-name)))
+                   (make-ement-session :user user :server server :transaction-id transaction-id
+                                       :device-id device-id :initial-device-display-name initial-device-display-name
+                                       :events (make-hash-table :test #'equal))))
+                (password-login
+                 () (pcase-let* (((cl-struct ement-session user device-id initial-device-display-name) session)
+                                 ((cl-struct ement-user id) user)
+                                 (data (ement-alist "type" "m.login.password"
+                                                    "identifier"
+                                                    (ement-alist "type" "m.id.user"
+                                                                 "user" id)
+                                                    "password" (or password
+                                                                   (read-passwd (format "Password for %s: " id)))
+                                                    "device_id" device-id
+                                                    "initial_device_display_name" initial-device-display-name)))
+                      ;; TODO: Clear password in callback (if we decide to hold on to it for retrying login timeouts).
+                      (ement-api session "login" :method 'post :data (json-encode data)
+                        :then (apply-partially #'ement--login-callback session))
+                      (ement-message "Logging in with password...")))
+                (sso-filter
+                 (process string)
+                 ;; NOTE: This is technically wrong, because it's not guaranteed that the
+                 ;; string will be a complete request--it could just be a chunk.  But in
+                 ;; practice, if this works, it's much simpler than setting up process log
+                 ;; functions and per-client buffers for this throwaway, pretend HTTP server.
+                 (when (string-match (rx "GET /?loginToken=" (group (0+ nonl)) " " (0+ nonl)) string)
+                   (unwind-protect
+                       (pcase-let* ((token (match-string 1 string))
+                                    ((cl-struct ement-session user device-id initial-device-display-name)
+                                     session)
+                                    ((cl-struct ement-user id) user)
+                                    (data (ement-alist
+                                           "type" "m.login.token"
+                                           "identifier" (ement-alist "type" "m.id.user"
+                                                                     "user" id)
+                                           "token" token
+                                           "device_id" device-id
+                                           "initial_device_display_name" initial-device-display-name)))
+                         (ement-api session "login" :method 'post
+                           :data (json-encode data)
+                           :then (apply-partially #'ement--login-callback session)))
+                     (delete-process sso-server-process)
+                     (delete-process process))))
+                (sso-login ()
+                           (setf sso-server-process
+                                 (make-network-process
+                                  :name "ement-sso" :family 'ipv4 :host 'local :service ement-sso-server-port
+                                  :filter #'sso-filter :server t :noquery t))
+                           (browse-url
+                            (concat (ement-server-uri-prefix (ement-session-server session))
+                                    "/_matrix/client/r0/login/sso/redirect?redirectUrl=http://localhost:"
+                                    (number-to-string ement-sso-server-port))))
+                (flows-callback
+                 (data) (let ((flows (cl-loop for flow across (map-elt data 'flows)
+                                              collect (map-elt flow 'type))))
+                          (pcase (length flows)
+                            (1 (pcase (car flows)
+                                 ("m.login.password" (password-login))
+                                 ("m.login.sso" (sso-login))
+                                 (_ (error "Ement: Unsupported login flow: %s  Server:%s"
+                                           (car flows) (ement-server-uri-prefix (ement-session-server session))))))
+                            (_ (pcase (completing-read "Select authentication method: "
+                                                       (cl-loop for flow in flows
+                                                                collect (string-trim-left flow (rx "m.login."))))
+                                 ("password" (password-login))
+                                 ("sso" (sso-login))
+                                 (else (error "Ement: Unsupported login flow:%S  Server:%S  Supported flows:%S"
+                                              else (ement-server-uri-prefix (ement-session-server session)) flows))))))))
+      (if session
+          ;; Start syncing given session.
+          (let ((user-id (ement-user-id (ement-session-user session))))
+            ;; HACK: If session is already in ement-sessions, this replaces it.  I think that's okay...
+            (setf (alist-get user-id ement-sessions nil nil #'equal) session)
+            (ement--sync session :timeout ement-initial-sync-timeout))
+        ;; Start password login flow.  Prompt for user ID and password
+        ;; if not given (i.e. if not called interactively.)
+        (unless user-id
+          (setf user-id (read-string "User ID: ")))
+        (setf session (new-session))
+        (when (ement-api session "login" :then #'flows-callback)
+          (message "Ement: Checking server's login flows..."))))))
 
 (defun ement-disconnect (sessions)
   "Disconnect from SESSIONS.
@@ -389,9 +439,12 @@ If no URI is found, prompt the user for the hostname."
                       ("" hostname)
                       (_ input))))
               (parse (string)
-                     (if-let ((object (ignore-errors (json-read-from-string string))))
-                         ;; Return extracted value.
-                         (map-nested-elt object '(m.homeserver base_url))
+                     (if-let* ((object (ignore-errors (json-read-from-string string)))
+                               (url (map-nested-elt object '(m.homeserver base_url)))
+                               ((string-match-p
+                                 (rx bos "http" (optional "s") "://" (1+ nonl))
+                                 url)))
+                         url
                        ;; Parsing error: FAIL_PROMPT.
                        (fail-prompt))))
     (condition-case err
@@ -461,7 +514,7 @@ a filter ID).  When unspecified, the value of
                                              (pcase (plz-response-status response)
                                                ((or 429 502) (setf reason "failed")))))
                                           ((pcase curl-error
-                                             (28 (setf reason "timed out")))))
+                                             (`(28 . ,_) (setf reason "timed out")))))
                                     (if reason
                                         (if (not ement-auto-sync)
                                             (run-hook-with-args 'ement-interrupted-sync-hook session)
@@ -868,16 +921,22 @@ and `session' to the session.  Adds function to
                                                 :ascent 'center
                                                 :max-width ement-room-avatar-max-width
                                                 :max-height ement-room-avatar-max-height)))
-                       (when (fboundp 'imagemagick-types)
-                         ;; Only do this when ImageMagick is supported.
-                         ;; FIXME: When requiring Emacs 27+, remove this (I guess?).
-                         (setf (image-property image :type) 'imagemagick))
-                       ;; We set the room-avatar slot to a propertized string that
-                       ;; displays as the image.  This seems the most convenient thing to
-                       ;; do.  We also unset the cached room-list-avatar so it can be
-                       ;; remade.
-                       (setf (ement-room-avatar room) (propertize " " 'display image)
-                             (alist-get 'room-list-avatar (ement-room-local room)) nil))))))
+                       (if (not image)
+                           (progn
+                             (display-warning 'ement (format "Room avatar seems unreadable:  ROOM-ID:%S  AVATAR-URL:%S"
+                                                             (ement-room-id room) (ement--mxc-to-url url session)))
+                             (setf (ement-room-avatar room) nil
+                                   (alist-get 'room-list-avatar (ement-room-local room)) nil))
+                         (when (fboundp 'imagemagick-types)
+                           ;; Only do this when ImageMagick is supported.
+                           ;; FIXME: When requiring Emacs 27+, remove this (I guess?).
+                           (setf (image-property image :type) 'imagemagick))
+                         ;; We set the room-avatar slot to a propertized string that
+                         ;; displays as the image.  This seems the most convenient thing to
+                         ;; do.  We also unset the cached room-list-avatar so it can be
+                         ;; remade.
+                         (setf (ement-room-avatar room) (propertize " " 'display image)
+                               (alist-get 'room-list-avatar (ement-room-local room)) nil)))))))
       ;; Unset avatar.
       (setf (ement-room-avatar room) nil
             (alist-get 'room-list-avatar (ement-room-local room)) nil))))
